class APIException(Exception):
    status_code = 400
    def __init__(self, message, status_code=None, payload=None) -> None:
        super().__init__(f"[{status_code}] {message}")

        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv


import json
import decimal
import datetime
from json.encoder import JSONEncoder
from typing import Tuple
from flask import Response
from flask.wrappers import Request

class CustomJSONDecoder(json.JSONEncoder):
    """ Clase que ayuda con el manejo de JSON de un blob Storage de Azure
    """
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            if o % 1 > 0:
                return float(o)
            else:
                return int(o)
        if isinstance(o, bytes):
            return o.decode()
        if isinstance(o, (datetime.date, datetime.datetime)):
            return o.isoformat()
        return super(CustomJSONDecoder, self).default(o)

def build_response(status: int, body: dict, application_type: str = 'application/json', jsonEncoder: JSONEncoder = CustomJSONDecoder, circular: bool = True, is_body_str: bool = False, encoder_extras: dict = {}) -> Response:
    """ Devuelve el formato que acepta azure para una respuesta de HTTP

    Args:
        status (int): Codido http
        body (dict): Contenido del json de respuesta
        application_type (str, optional): Tipo de respuesta. Defaults to 'application/json'.
        jsonEncoder (JSONEncoder, optional): Codificacion el JSON de salida. Defaults to CustomJSONDecoder.
        circular (bool, optional): Inidica si la codificacion la hara por cada parametro del JSON. Defaults to True.

    Returns:
        func.HttpResponse: Respuesta HTTP aceptada por Azure
    """
    return Response(
        status=status,
        response=body if is_body_str else json.dumps(body, cls=jsonEncoder, check_circular=circular, **encoder_extras),
        mimetype=application_type
    )

def serialize_json(data: dict, jsonEncoder: JSONEncoder = CustomJSONDecoder, circular: bool = True) -> str:
    """ Devuelve una cadena en formato JSON de un objeto

    Args:
        data (dict): Contenido del json de respuesta
        jsonEncoder (JSONEncoder, optional): Codificacion el JSON de salida. Defaults to CustomJSONDecoder.
        circular (bool, optional): Inidica si la codificacion la hara por cada parametro del JSON. Defaults to True.

    Returns:
        str: Cadena con formato JSON del contenido
    """
    return json.dumps(data, cls=jsonEncoder, check_circular=circular)

def get_paginate_params(req: Request) -> Tuple[bool, int, int]:
    """ Devuelve los parametros de paginacion de una peticion http

    Args:
        req (func.HttpRequest): Peticion http

    Returns:
        Tuple[bool, int, int]: Parametros de paginacion (Paginado, num de pagina, elementos por pagina)
    """

    page = req.args.get('page')
    if page is not None:
        page = int(page)
    else:
        page = 1

    per_page = req.args.get('per_page')
    if per_page is not None:
        per_page = int(per_page)
    else:
        per_page = 10
    
    return (page, per_page)

def get_filter_params(req: Request) -> dict:
    """ Obtiene filtros de query

    Args:
        req (dict): Peticion http

    Returns:
        dict: Filtros formados como par valor
    """
    if req.args is None:
        return {}
    
    ret_dict = {}
    for key in req.args.keys():
        if key == 'page' or key == 'per_page' or key == 'relationships':
            pass
        else:
            ret_dict.update({key: req.args[key]})
    
    return ret_dict

def get_relationship_params(req: Request) -> dict:
    """ Obtiene filtros de query

    Args:
        req (dict): Peticion http

    Returns:
        dict: Filtros formados como par valor
    """
    if req.args is None:
        return {}
    
    ret_dict = {}
    if 'relationships' in req.args.keys():
        ret_dict.update(dict(relationships=req.args.getlist('relationships')))
    
    return ret_dict
 
def get_search_params(req: Request) -> dict:
    """ Obtiene filtros de query

    Args:
        req (dict): Peticion http

    Returns:
        dict: Filtros formados como par valor
    """
    if req.args is None:
        return {}
    
    ret_dict = {}
    for k in req.args.keys():
            
        if str(k).startswith('search-'):
            key = str(k).split('-')[1]
            ret_dict[key] = str(req.args[k]).replace('*', '%')
    
    return ret_dict
 
def get_search_method_param(req: Request) -> str:
    """ Obtiene el metodo de filtrado de query

    Args:
        req (dict): Peticion http

    Returns:
        str: Metodo de filtraddo
    """
    if req.args is None:
        return 'AND'
    if 'searchmethod' not in req.args:
        return 'AND'
    
    method = str(req.args['searchmethod']).upper()
    return 'AND' if method not in ['AND', 'OR'] else method



import re
from typing import Any, Match, Type
from sqlalchemy.orm.session import Session
from flask import request
from sqlalchemy.sql.elements import literal
from sqlalchemy.sql.schema import Column

from ..Exceptions.APIException import APIException
from ..Data.Enum.request_parts import RequestPart

class DBValidator:
    def __init__(self, type: str, table: Type, column: Column[int]) -> None:
        self.type = type
        self.table = table
        self.column = column
        pass

class RequestValidator():
    """ Validador de peticiones HTTP entrantes
    """

    request = None
    rules = None
    req_part = None
    is_valid = True
    errors = None
    error_code = 422
    session = None

    def __init__(self, session: Session, rules: dict, req_part: str = 'body',):
        self.rules = rules
        self.req_part = req_part
        self.session = session

    def validate(self) -> bool:
        request_parts = [req.value for req in RequestPart]
        if self.rules is None:
            self.is_valid = True
            raise APIException("Validation not pass", status_code=self.error_code, payload=self.errors)
        if not self.req_part in request_parts:
            self.is_valid = False
            self.errors = {"error": f"The section is not in Request Part. [{str(', ').join(request_parts)}]", "field":"req_part"}
            self.error_code = 500
            raise APIException("Validation not pass", status_code=self.error_code, payload=self.errors)
        "Validate all the rules defined with the request"
        try:
            if self.req_part == RequestPart.BODY.value:
                self.request = request.get_json()
            if self.req_part == RequestPart.PARAM.value:
                self.request = request.view_args
            if self.req_part == RequestPart.QUERY.value:
                self.request = request.args
        except ValueError:
            self.is_valid = False
            self.errors = {"error": "Can't proccess the request", "field": self.req_part}
            self.error_code = 422
            raise APIException("Can't proccess the request", status_code=self.error_code, payload=self.errors)

        self.is_valid = True
        for field in self.rules:
            is_none = False
            request_value = self.request.get(field)
            rule_params = self.rules.get(field)
            for rule_param in rule_params:
                if rule_param == 'nullable' and request_value is None:
                    is_none = True
                    break
                if rule_param == 'required' and not field in self.request:
                    self.is_valid = False
                    self.errors = {"error": f"The field {field} is required", "field":field}
                    break
                elif (not is_none) and rule_param == 'string' and not isinstance(request_value, str):
                    self.is_valid = False
                    self.errors = {"error": f"The field {field} should be text", "field":field}
                    break
                elif (not is_none) and rule_param == 'boolean' and not isinstance(request_value, bool):
                    self.is_valid = False
                    self.errors = {"error": f"The field {field} should be true/false", "field":field}
                    break
                elif (not is_none) and rule_param == 'numeric' and not isinstance(request_value, (int, float)) and not request_value.isdigit():
                    self.is_valid = False
                    self.errors = {"error": f"The field {field} should be a number", "field":field}
                    break
                elif (not is_none) and rule_param == 'email' and not self.is_mail(request_value):
                    self.is_valid = False
                    self.errors = {"error": f"The field {field} should be a valid email", "field":field}
                    break
                elif (not is_none) and isinstance(rule_param, DBValidator) and rule_param.type =='exists':
                    if rule_param.table is None or rule_param.column is None:
                        self.is_valid = False
                        self.errors = {"error": "Validator format error", "field":field}
                        self.error_code = 500
                        break
                    if not self.exists(rule_param, request_value):
                        self.is_valid = False
                        self.errors = {"error": f"The value {request_value} doesn't exits in the column {rule_param.column} of the table {rule_param.table}", "field":field}
                        self.error_code = 422
                        break
                elif (not is_none) and isinstance(rule_param, DBValidator) and rule_param.type == 'unique':
                    if rule_param.table is None or rule_param.column is None:
                        self.is_valid = False
                        self.errors = {"error": "Validator format error", "field":field}
                        self.error_code = 500
                        break
                    if self.exists(rule_param, request_value):
                        self.is_valid = False
                        self.errors = {"error": f"The {request_value} already exists in {rule_param.column} of the table {rule_param.table}", "field":field}
                        self.error_code = 422
                        break
            if not self.is_valid:
                break

        if not self.is_valid:
            raise APIException("Can't proccess the request", status_code=self.error_code, payload=self.errors)
        
        return True

    def is_mail(self, text: str) -> Match | None:
        """ Verify if the input string has a email format

        Args:
            text (str): Email string

        Returns:
            Match or None: Indicates if there is a match according with email regex
        """
        email_regex = '^(\w|\.|\_|\-)+[@](\w|\_|\-|\.)+[.]\w{2,3}$'
        return re.search(email_regex, text)
    
    def exists(self, database_params: DBValidator, value: Any) -> bool:
        """ Define if a field exists in database

        Args:
            database_params (dict): Evaluated parameter
            value (Any): Value to verify

        Returns:
            bool: indicate if the value exists
        """
        return self.session.query(literal(True)).filter(database_params.column == value).first() is not None
    


import json
import decimal
import datetime
from typing import List
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm.session import Session as ORMSession

## Database connection string
db: SQLAlchemy = SQLAlchemy()


def get_session() -> ORMSession:
    """ Return a new database session from engine to data access

    Returns:
        ORMSession: Database session
    """
    return db.session

def get_engine() -> Engine:
    """ Return the database engine

    Returns:
        Engine: Database Engines
    """
    return db.engine

def connect_to_main_app(db_conn: dict):
    """
    Connects the application to the specified database connection.

    Args:
        db_conn (dict): A dictionary containing the database connection details.

    Returns:
        None: This function does not return a value, but updates the global 'db' variable with the provided database connection.
    """
    global db
    db = db_conn['db']


class AlchemyEncoder(json.JSONEncoder):
    """ Based on: https://stackoverflow.com/questions/5022066/how-to-serialize-sqlalchemy-result-to-json/41204271 """
    def default(self, obj):
        if isinstance(obj.__class__, DeclarativeMeta):
            fields = {}
            prop_map_obj = obj.__class__.property_map()
            for field in [x for x in obj.attrs]:
                data = obj.__getattribute__(field)
                try:
                    if isinstance(data, (datetime.datetime, datetime.date, datetime.time)):
                        data = data.isoformat()
                    else:
                        json.dumps(data)
                    fields[prop_map_obj[field] if field in prop_map_obj else field] = data
                except TypeError:
                    fields[field] = None
            return fields
        if isinstance(obj, decimal.Decimal):
            if obj % 1 > 0:
                return float(obj)
            else:
                return int(obj)
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


class AlchemyRelationEncoder(json.JSONEncoder):
    def __init__(self, relationships: List[str], **kwargs) -> None:
        super().__init__(**kwargs)
        self.relationships = relationships
        
    """ Based on: https://stackoverflow.com/questions/5022066/how-to-serialize-sqlalchemy-result-to-json/41204271 """
    def default(self, obj):
        if isinstance(obj.__class__, DeclarativeMeta):
            fields = {}
            prop_map_obj = obj.__class__.property_map()
            
            relation_names = [attr for attr, relation in obj.__mapper__.relationships.items()]
            filters_model = list(set(self.relationships).intersection(relation_names))
            attributes = [x for x in obj.attrs]
            
            if type(filters_model) is list:
                attributes.extend(filters_model)
            
            for field in attributes:
                data = obj.__getattribute__(field)
                try:
                    if isinstance(data, (datetime.datetime, datetime.date, datetime.time)):
                        data = data.isoformat()
                    else:
                        json.dumps(data, cls=self.__class__, check_circular=self.check_circular, relationships=self.relationships)
                    fields[prop_map_obj[field] if field in prop_map_obj else field] = data
                except TypeError as e:
                    fields[field] = None
            return fields
        if isinstance(obj, decimal.Decimal):
            if obj % 1 > 0:
                return float(obj)
            else:
                return int(obj)
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)




import os
from functools import wraps
from typing import Any, Callable

def env(env_key: str, default_value: Any) -> Any:
    """ Parsea el valor de una variable de entorno a una variable utilizable para python

    Args:
        env_key (str): Nombre de la variable de entorno de
        default_value (Any): Valor por default de la variable de entorno

    Returns:
        Any: Valor designado de la variable de entorno o en su defecto la default
    """
    if env_key in os.environ:
        if os.environ[env_key].isdecimal():
            return int(os.environ[env_key])
        elif str(os.environ[env_key]).lower() == "true" or str(os.environ[env_key]).lower() == "true":
            return str(os.environ[env_key]).lower() == "true"
        else:
            return os.environ[env_key]
    else:
        return default_value

def reload_env_vars(function: Callable):
    @wraps(function)
    def decorator(*args, **kwargs):
        global APP_URL
        APP_URL     = env("APP_URL", "")
        
        function(*args, **kwargs)

    return decorator

APP_URL     = env("APP_URL", "")




from typing import List, Tuple, Type, cast

from sqlalchemy.orm.query import Query
from ..Data.BaseModel import BaseModel
from sqlalchemy.orm.session import Session

class BaseService:
    def __init__(self, model: Type) -> None:
        self.model = model
    
    def get_all(self, session: Session, paginate = False, page = 1, per_page = 10) -> Tuple[Query, BaseModel]:
        """ Obtiene todos los elementos del modelo de datos especificado

        Args:
            self (class): Class
            session (Session): Database session
            paginate (bool, optional): Flag to paginate results. Defaults to False.
            page (int, optional): Pagenumber to return. Defaults to 1.
            per_page (int, optional): Number of elements per page. Defaults to 10.

        Returns:
            List: Lista de objectos de base de datos
        """
        if paginate is True:
            return cast(BaseModel, self.model).get_paginated(session, page, per_page)
        return cast(BaseModel, self.model).all(session)
    
    def get_one(self, session: Session, id: int):
        """ Search an element by id

        Args:
            self (class): Class
            session (Session): Database session
            id (int): Database identifier

        Returns:
            ORMClass: Devuelve un objeto de la base de datos
        """
        return cast(BaseModel, self.model).find(session, id)
    
    def filter_by_column(self, session: Session, column_name: str, column_value, paginate = False, page = 1, per_page = 10, first: bool = False):
        return cast(BaseModel, self.model).filter_by(session, column_name, column_value, paginate, page, per_page, first)

    def get_by_column(self, session: Session, column_name: str, column_value):
        """ Get an element according with the specified colum name and value

        Args:
            self (class): Class
            session (Session): Database session
            column_name (str): Column name
            value (Any): Value to mach

        Returns:
            ORMClass: First coincidence of the search
        """
        return cast(BaseModel, self.model).get_one(session, column_name, column_value)
    
    def multiple_filters(self, session: Session, filters: List[dict], paginate = False, page = 1, per_page = 10, first: bool = False, search_filters: dict = {}, search_method='AND'):
        return cast(BaseModel, self.model).filters(session, filters, paginate, page, per_page, first, search_filters, search_method)
    
    def count_with_query(self, query: Query) -> int:
        return query.count()
    
    def count_elements(self, session: Session) -> int:
        return cast(BaseModel, self.model).count(session)
    
    def count_filtered(self, session: Session, filters: List[dict]) -> int:
        return cast(BaseModel, self.model).count_with_filters(session, filters)
    
    def insert_register(self, session: Session, input_data: dict):
        input_params = {}
        for ipKey in input_data.keys():
            if ipKey in self.get_display_members():
                input_params[ipKey] = input_data[ipKey]
        obj = self.model(**input_params)
        return cast(BaseModel, obj).save(session, **input_data)
    
    def update_register(self, session: Session, id: int, update_data: dict):
        obj = self.get_one(session, id)
        return cast(BaseModel, obj).update(session, update_data)
    
    def delete_register(self, session: Session, id: int):
        obj = self.get_one(session, id)
        return cast(BaseModel, obj).delete(session)
    
    def get_rules_for_store(self):
        return cast(BaseModel, self.model).rules_for_store()
    
    def get_display_members(self) -> List[str]:
        return cast(BaseModel, self.model).display_members()

    def get_filter_columns(self) -> List[str]:
        return cast(BaseModel, self.model).filter_columns
    
    def get_search_columns(self) -> List[str]:
        return cast(BaseModel, self.model).search_columns

    def get_relationship_names(self) -> List[str]:
        return cast(BaseModel, self.model).relationship_names



from typing import Dict, Type
from ...Data.BaseModel import BaseModel
import isiflask_core.Environment as env

class ResourceReference:

    @env.reload_env_vars
    def __init__(self, model: Type[BaseModel], prefix_model: str = "", sufix_model: str = "", action: str = "GET", api_version = "v1", is_json_resp = True) -> None:
        self.Name = model.__name__
        self.Action = action
        self.Ref = f"{env.APP_URL}/api/{api_version}{prefix_model}/{model.model_path_name}{sufix_model}" if is_json_resp is True else f"{env.APP_URL}/{model.model_path_name}{sufix_model}"
    
    def to_dict(self) -> Dict:
        return {
            "Name": self.Name,
            "Action": self.Action,
            "Ref": self.Ref
        }


from enum import Enum

class HTTPStatusCode(Enum):
    """ Http status code enum class
    """

    OK  = 200
    CREATED = 201
    NO_CONTENT = 204

    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    FORBIDDEN = 403

    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406
    CONFLICT = 409
    UNSUPPORTED_TYPE = 415
    UNPROCESABLE_ENTITY = 422

    INTERNAL_SERVER_ERROR = 500


import json
from typing import cast
from flask import request
import logging
from ..Data.Enum.http_status_code import HTTPStatusCode

from ..Exceptions.APIException import APIException
from ..Validators.RequestValidator import RequestValidator

from ..Data.Interfaces.PaginationResult import PaginationResult
from ..Data.BaseModel import BaseModel
from ..Services.BaseService import BaseService
from ...database.DBConnection import AlchemyEncoder, AlchemyRelationEncoder, get_session
from ...utils.http_utils import build_response, get_paginate_params, get_filter_params, get_relationship_params, get_search_method_param, get_search_params

def index(service: BaseService):
    session = get_session()
    (page, per_page) = get_paginate_params(request)
    relationship_retrieve = get_relationship_params(request)
    filter_query = get_filter_params(request)
    filter_keys = filter_query.keys()
    
    encoder = AlchemyEncoder if 'relationships' not in relationship_retrieve else AlchemyRelationEncoder
    search_query = get_search_params(request)
    search_keys = service.get_search_columns()
    search_columns = list(set(search_keys).intersection(search_query.keys()))
    filters_search = []
    for skey in search_columns:
        filters_search.append({
            'column': getattr(cast(BaseService, service).model, skey),
            'value': search_query[skey]
        })
    
    search_method = 'AND'
    if len(filters_search) > 0:
        search_method = get_search_method_param(request)
    
    model_filter_keys = cast(BaseService, service).get_filter_columns()
    filters_model = set(model_filter_keys).intersection(filter_keys)
    
    filters = []
    for f in filters_model:
        filters.append({f: filter_query[f]})
    
    try:
        query, elements = cast(BaseService, service).multiple_filters(session, filters, True, page, per_page, search_filters=filters_search, search_method=search_method)
        total_elements = cast(BaseService, service).count_with_query(query)
        
        if 'accepts' in request.headers:
            accepts = request.headers['accepts']
            if accepts.lower() == 'application/json':
                body = PaginationResult(elements, page, per_page, total_elements, refType=cast(BaseService, service).model).to_dict()
            else:
                body = PaginationResult(elements, page, per_page, total_elements, refType=cast(BaseService, service).model, is_json_resp=False).to_dict()
        else:
            body = PaginationResult(elements, page, per_page, total_elements, refType=cast(BaseService, service).model, is_json_resp=False).to_dict()
        body['Data'] = list(map(lambda d: dict(
                **cast(BaseModel, d).to_dict(jsonEncoder=encoder, encoder_extras=relationship_retrieve)
            ), body['Data'])
        )

        status_code = HTTPStatusCode.OK.value
    except APIException as e:
        logging.exception("APIException occurred")
        body = e.to_dict()
        status_code = e.status_code
    except Exception as e:
        logging.exception("Cannot make the request")
        print(str(e))
        body = dict(message=str(e))
        status_code = HTTPStatusCode.UNPROCESABLE_ENTITY.value
    finally:
        session.close()
    
    return build_response(status_code, body, jsonEncoder=encoder, encoder_extras=relationship_retrieve)

def find(service: BaseService, id: int):
    session = get_session()
    relationship_retrieve = get_relationship_params(request)
    encoder = AlchemyEncoder if 'relationships' not in relationship_retrieve else AlchemyRelationEncoder
    try:
        element = cast(BaseService, service).get_one(session, id)
        body = element.to_dict(jsonEncoder=encoder, encoder_extras=relationship_retrieve)
        status_code = HTTPStatusCode.OK.value
    except APIException as e:
        logging.exception("APIException occurred")
        body = e.to_dict()
        status_code = e.status_code
    except Exception as e:
        logging.exception("Cannot make the request")
        body = dict(message="Cannot make the request")
        status_code = HTTPStatusCode.UNPROCESABLE_ENTITY.value
    finally:
        session.close()
    return build_response(status_code, body, jsonEncoder=AlchemyEncoder)

def store(service: BaseService):
    session = get_session()
    
    RequestValidator(session, cast(BaseService, service).get_rules_for_store()).validate()
    input_params = request.get_json()

    try:
        body = cast(BaseService, service).insert_register(session, input_params)
        response = json.dumps(body, cls=AlchemyEncoder)
        status_code = HTTPStatusCode.OK.value
    except APIException as e:
        logging.exception("APIException occurred")
        response = json.dumps(e.to_dict())
        status_code = e.status_code
    except Exception:
        logging.exception("No se pudo realizar la consulta")
        body = dict(message="No se pudo realizar la consulta")
        response = json.dumps(body)
        status_code=HTTPStatusCode.UNPROCESABLE_ENTITY.value
    finally:
        session.close()
    
    return build_response(status_code, response, is_body_str=True)

def update(service: BaseService, id: int):
    session = get_session()

    input_params = request.get_json()
    try:
        body = cast(BaseService, service).update_register(session, id, input_params)
        response = json.dumps(body, cls=AlchemyEncoder)
        status_code = HTTPStatusCode.OK.value
    except APIException as e:
        logging.exception("APIException occurred")
        response = json.dumps(e.to_dict())
        status_code = e.status_code
    except Exception as e:
        logging.exception("Cannot make the request")
        body = dict(message="Cannot make the request")
        response = json.dumps(body)
        status_code = HTTPStatusCode.UNPROCESABLE_ENTITY.value
    finally:
        session.close()
    return build_response(status_code, response, is_body_str=True)

def delete(service: BaseService, id: int):
    session = get_session()

    try:
        body = cast(BaseService, service).delete_register(session, id)
        status_code = HTTPStatusCode.NO_CONTENT.value
    except APIException as e:
        logging.exception("APIException occurred")
        body = e.to_dict()
        status_code = e.status_code
    except Exception as e:
        logging.exception("Cannot make the request")
        body = dict(message="Cannot make the request")
        status_code = HTTPStatusCode.UNPROCESABLE_ENTITY.value
    finally:
        session.close()
    return build_response(status_code, body, jsonEncoder=AlchemyEncoder)


from typing import List, Type
from flask import request

from ...Data.BaseModel import BaseModel
from .ResourceReference import ResourceReference

class PaginationResult:

    def __init__(self, data: List[BaseModel], offset: int = 1, limit: int = 1, total: int = 1, prefix_model: str = "", sufix_model: str = "", refType: Type = BaseModel, is_json_resp = True) -> None:
        if len(data) > 0:
            self.Data = data
            #self.Links = {
            #    "next": ResourceReference(type(data[0]), prefix_model='/catalog').to_dict(),
            #    "current": ResourceReference(type(data[0]), prefix_model='/catalog').to_dict(),
            #    "prev": ResourceReference(type(data[0]), prefix_model='/catalog').to_dict()
            #}
            self.Links = {
                "current": ResourceReference(
                    type(data[0]),
                    prefix_model=prefix_model,
                    sufix_model=f"{'/' if sufix_model != '' else ''}{sufix_model}?page={offset}&per_page={limit}",
                    action=request.method,
                    is_json_resp=is_json_resp)
                    .to_dict()
            }
            self.Offset = offset
            self.Limit = limit
            self.Total = total

            if (self.Offset * self.Limit) < self.Total:
                self.Links["next"] = ResourceReference(
                    refType,
                    prefix_model=prefix_model,
                    sufix_model=f"{'/' if sufix_model != '' else ''}{sufix_model}?page={offset+1}&per_page={limit}",
                    action=request.method,
                    is_json_resp=is_json_resp).to_dict()
            if self.Offset > 1:
                self.Links["prev"] = ResourceReference(
                    refType,
                    prefix_model=prefix_model,
                    sufix_model=f"{'/' if sufix_model != '' else ''}{sufix_model}?page={offset-1}&per_page={limit}",
                    action=request.method,
                    is_json_resp=is_json_resp).to_dict()
        else:
            self.Data = []
            self.Links = None
            self.Offset = 0
            self.Limit = 0
            self.Total = 0
    
    def to_dict(self) -> dict:
        return {
            "Data": self.Data,
            "Links": self.Links,
            "Offset": self.Offset,
            "Limit": self.Limit,
            "Total": self.Total
        }
    


from __future__ import annotations
import json
from json.encoder import JSONEncoder
from typing import Any, Dict, List, Type, cast
from operator import and_, or_
from sqlalchemy import Column, Integer, orm
from sqlalchemy.orm.session import Session
from sqlalchemy.orm.query import Query
from typing import List

from ...database.DBConnection import db, AlchemyEncoder

class BaseModel(db.Model):
    """ Base model for a child classes implementations

    Args:
        Base (Any): SQLAlchemy declarative base

    Raises:
        e: Exceptions for not implemented functions

    Returns:
        BaseModel: Instace of BaseModel class
    """

    ## Indicate it is an abstract class (not 100% completed y going to be done in child classes)
    __abstract__ = True
    ## Model identifier. All tables from database should have
    id = Column(Integer, primary_key=True, autoincrement=True)
    model_path_name = ""
    
    filter_columns = []
    relationship_names = []
    search_columns = []

    @property
    def attrs(self) -> List[str]:
        """ Returns a list of the attributes of an object

        Returns:
            List[str]: Attributes
        """
        preliminar = list(filter(lambda prop: not str(prop).startswith('_'), type(self).__dict__.keys()))
        display_member = self.__class__.display_members()
        return list(set(preliminar) & set(display_member)) if len(display_member) > 0 else display_member

    @classmethod
    def all(cls_, session: Session):
        """ Get all rows from a table

        Args:
            cls_ (cls): Type of class
            session (Session): Database session

        Returns:
            List[Type[BaseModel]]: List of elements mapped from database table
        """
        query = session.query(cls_)
        return query, session.query(cls_).all()
    
    @classmethod
    def get_paginated(cls_, session: Session, page: int = 1, per_page: int = 10):
        page = page - 1
        query = session.query(cls_).order_by(cls_.id.desc()).limit(per_page).offset(page*per_page)
        return query.all()
    
    @classmethod
    def find(cls_, session: Session, id: int):
        """ Search a row by id

        Args:
            cls_ (class): Child class method
            session (Session): Database session
            id (int): Row identifier

        Returns:
            Type[BaseModel]: The row that have a coincidence with the identifier
        """
        if int(id) > 0:
            return session.query(cls_).get(id)
    
    @classmethod
    def filter_by(cls_, session: Session, column_name: str, value, paginated: bool = False, page: int = 1, per_page: int = 10, first = False):
        """ Gets all rows that match with the specified filter

        Args:
            cls_ (class): Child class method
            session (Session): Database session
            column_name (str): Column name to filter
            value (Any): Value to match

        Returns:
            List[Type[BaseModel]]: List of elements that match with filter
        """
        filter_dict = {
            column_name: value
        }
        query = session.query(cls_).filter_by(**filter_dict).order_by(cls_.id.desc())

        if first:
            return query, query.first()

        if paginated:
            page = page - 1
            return query, query.limit(per_page).offset(page*per_page).all()
        return query, query.all()
    
    @classmethod
    def get_one(cls_, session: Session, column_name: str, value):
        """ Gets the first row that matches the filter

        Args:
            cls_ (class): Child class method
            session (Session): Database session
            column_name (str): Column name to filter
            value (Any): Value to match

        Returns:
            Type[BaseModel]: First register that match with filter
        """
        filter_dict = {
            column_name: value
        }
        return session.query(cls_).filter_by(**filter_dict).first()
    
    @classmethod
    def filters(cls_, session: Session, filters: List[dict], paginated: bool = False, page: int = 1, per_page: int = 10, first: bool = False, search_filters: dict = {}, search_method = 'AND'):
        """ Gets all rows that match with the multiple filters specified in dict (and logic)

        Args:
            cls_ (class): Child class method
            session (Session): Database session

        Returns:
            List[Type[BaseModel]]: List of elements that match with the multiple filters
        """
        query = session.query(cls_)
        
        search_query = None
        first_run = True
        for ksearch in search_filters:
            if first_run:
                search_query = cast(Column, ksearch['column']).ilike(ksearch['value'])
                first_run = False
            else:
                if search_method == 'OR':
                    search_query = or_(search_query, cast(Column, ksearch['column']).ilike(ksearch['value']))
                else:
                    search_query = and_(search_query, cast(Column, ksearch['column']).ilike(ksearch['value']))
        
        if search_query is not None:
            query = query.filter(search_query)

        for filter in filters:
            query = query.filter_by(**filter)
        
        query = query.order_by(cls_.id.desc())

        if first:
            return query, query.first()
        
        if paginated:
            page = page - 1
            return query, query.limit(per_page).offset(page*per_page).all()
        return query, query.all()

    def before_save(self, sesion: Session, *args, **kwargs):
        """ Method to execute before save a row in database (polimorfism)
        """
        pass

    def after_save(self, sesion: Session, *args, **kwargs):
        """ Method to execute after save a row in database (polimorfismo)
        """
        pass
    
    def save(self, session: Session, commit=True, *args, **kwargs):
        """ Save a register in database

        Args:
            session (Session): Database session
            commit (bool, optional): Indicate if the changes will make in database. Defaults to True.

        Raises:
            e: In case of error, the register will be erased and raise an Exception
        """
        self.before_save(session, *args, **kwargs)
        session.add(self)
        if commit:
            try:
                session.commit()
            except Exception as e:
                session.rollback()
                raise e

        self.after_save(session, *args, **kwargs)
        return self

    def before_update(self, sesion: Session, *args, **kwargs):
        """ Method to execute before update a row in database (polimorfism)
        """
        pass

    def after_update(self, sesion: Session, *args, **kwargs):
        """ Method to execute after update a row in database (polimorfism)
        """
        pass

    def update(self, session: Session, object: dict, *args, **kwargs):
        """ Update a specified register in database

        Args:
            session (Session): Database session
            object (dict): Dictionary with only the field to update
        """
        self.before_update(session, *args, **kwargs, **object)
        keys = self.__class__.get_keys()
        for key in keys:
            if key in object:
                self.__setattr__(key, object[key])
                pass

        session.commit()
        self.after_update(session, *args, **kwargs)
        return self
    
    def before_delete(self, sesion: Session, *args, **kwargs):
        """ Method to execute before update a row in database (polimorfism)
        """
        pass

    def after_delete(self, sesion: Session, *args, **kwargs):
        """ Method to execute after update a row in database (polimorfism)
        """
        pass

    def delete(self, session: Session, commit=True, *args, **kwargs):
        """ Delete a specified register in database

        Args:
            session (Session): Database session
            commit (bool, optional): Indicate if the changes will make in database. Defaults to True.
        """
        self.before_delete(session, *args, **kwargs)
        session.delete(self)
        if commit:
            session.commit()
        self.after_delete(session, *args, **kwargs)

    @classmethod
    def eager(cls_: Type[BaseModel], session: Session, *args) -> Query:
        """ Execute in one load all joins

        Returns:
            Type[BaseModel]: Database query
        """
        cols = [orm.joinedload(arg) for arg in args]
        return session.query(cls_).options(*cols)
    
    @classmethod
    def count(cls_: Type[BaseModel], session: Session) -> int:
        """ Execute in one load all joins

        Returns:
            Type[BaseModel]: Database query
        """
        return session.query(cls_.id).count()
    
    @classmethod
    def count_with_filters(cls_: Type[BaseModel], session: Session, filters: List[dict]) -> int:
        query = session.query(cls_)

        for filter in filters:
            query = query.filter_by(**filter)
        
        return query.count()
    
    @classmethod
    def get_keys(cls_: Type[BaseModel]) -> List[str]:
        """ Get all attributes of class

        Args:
            cls_ (Type[BaseModel]): Child class method

        Returns:
            List[str]:  Attributes
        """
        return list(filter(lambda prop: not str(prop).startswith('_'), cls_.__dict__.keys()))
    
    @classmethod
    def rules_for_store(cls_) -> Dict[str, List[Any]]:
        """Define a dictionary with the rules for each property defined

        Returns:
            Dict[str, List[Any]]: List of rules for each property
        """
        return {}
    
    @classmethod
    def property_map(cls_) -> Dict[str, str]:
        """Remap property with display value

        Returns:
            Dict[str, str]: Dict of string with key as class property name and value as display
        """
        return {}
    
    @classmethod
    def display_members(cls_) -> List[str]:
        """Get only de properties to display to end user

        Returns:
            List[str]: List of properties
        """
        return []
    
    def to_dict(self, jsonEncoder: JSONEncoder = AlchemyEncoder, circular: bool = True, encoder_extras: dict = {}) -> dict:
        return json.loads(json.dumps(self, cls=jsonEncoder, check_circular=circular, **encoder_extras))


    def __repr__(self) -> str:
        """ Model representation

        Returns:
            str: Model output string formatted
        """
        attr_array = [f"{attr}={self.__getattribute__(attr)}" for attr in self.attrs]
        args_format = ",".join(attr_array)
        return f"<{type(self).__name__}({args_format})>"


